==> incenp.binseqs-0.1.0.tar.gz: incenp.binseqs-0.1.0/incenp/bio/seqio/XdnaIO.py <==
# -*- coding: utf-8 -*-
# Copyright © 2017,2018 Damien Goutte-Gattat
#
# Redistribution and use of this script, with or without modifications,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of this script must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Xdna format support for Biopython

Provide read/write support for the binary format used by DNA Strider
and SerialCloner.
"""

from re import match
from struct import pack, unpack

from Bio import Alphabet
from Bio.Seq import Seq
from Bio.SeqIO.Interfaces import SequenceWriter
from Bio.SeqFeature import SeqFeature, FeatureLocation, ExactPosition
from Bio.SeqRecord import SeqRecord


_seq_types = {
    0: Alphabet.generic_alphabet,
    1: Alphabet.generic_dna,
    2: Alphabet.generic_dna,
    3: Alphabet.generic_rna,
    4: Alphabet.generic_protein
}

_seq_topologies = {
    0: 'linear',
    1: 'circular'
}


def _read(handle, length):
    data = handle.read(length)
    if len(data) < length:
        raise ValueError("Cannot read %d bytes from handle" % length)
    return data


def _read_pstring(handle):
    length = unpack('>B', _read(handle, 1))[0]
    return unpack('%ds' % length, _read(handle, length))[0].decode('ASCII')


def _read_pstring_as_integer(handle):
    return int(_read_pstring(handle))


def _read_overhang(handle):
    # An overhang is represented in a XDNA file as:
    # - a Pascal string containing the text representation of the
    #   overhang length (3' overhang if negative, 5' overhang if
    #   positive, no overhang if zero);
    # - the actual overhang sequence.
    length = _read_pstring_as_integer(handle)
    if length != 0:
        overhang = _read(handle, abs(length))
        return (length, overhang)
    else:
        return (None, None)


def _parse_feature_description(desc, qualifiers):
    # The 'description' field of a feature sometimes contains
    # several GenBank-like qualifiers, separated by CR characters.
    for part in [x for x in desc.split('\x0D') if len(x) > 0]:
        m = match('^([^=]+)="([^"]+)"?$', part)
        if m:
            qual, value = m.groups()
            qualifiers[qual] = [value]
        elif not '"' in part:   # Reject ill-formed qualifiers
            qualifiers['note'] = [part]


def _read_feature(handle, record):
    name = _read_pstring(handle)
    desc = _read_pstring(handle)
    type = _read_pstring(handle) or 'misc_feature'
    start = _read_pstring_as_integer(handle)
    end = _read_pstring_as_integer(handle)

    # Feature flags (4 bytes):
    # byte 1 is the strand (0: reverse strand, 1: forward strand);
    # byte 2 tells whether to display the feature;
    # byte 4 tells whether to draw an arrow when displaying the feature;
    # meaning of byte 3 is unknown.
    (forward, display, arrow) = unpack('>BBxB', _read(handle, 4))
    if forward:
        strand = 1
    else:
        strand = -1
        start, end = end, start

    # The last field is a Pascal string usually containing a
    # comma-separated triplet of numbers ranging from 0 to 255.
    # I suspect this represents the RGB color to use when displaying
    # the feature. Skip it as we have no need for it.
    _read_pstring(handle)

    # Assemble the feature
    # Shift start by -1 as XDNA feature coordinates are 1-based
    # while Biopython uses 0-based couting.
    location = FeatureLocation(start - 1, end, strand=strand)
    qualifiers = {}
    if name:
        qualifiers['label'] = [name]
    _parse_feature_description(desc, qualifiers)
    feature = SeqFeature(location, type=type, qualifiers=qualifiers)
    record.features.append(feature)


def XdnaIterator(handle):

    # Parse fixed-size header and do some rudimentary checks
    header = _read(handle, 112)
    (version, type, topology, length, neg_length, com_length) = \
            unpack('>BBB25xII60xI12x', header)
    if version != 0:
        raise ValueError("Unsupported XDNA version")
    if not type in _seq_types:
        raise ValueError("Unknown sequence type")

    # Read actual sequence and comment found in all XDNA files
    sequence = _read(handle, length).decode('ASCII')
    comment = _read(handle, com_length).decode('ASCII')

    # Try to derive a name from the first "word" of the comment.
    name = comment.split(' ', 2)[0]
    if len(name) > 16:
        name = None

    # Create record object
    record = SeqRecord(Seq(sequence, _seq_types[type]),
            description=comment, name=name)
    if topology in _seq_topologies:
        record.annotations['topology'] = _seq_topologies[topology]

    if len(handle.read(1)) == 1:
        # This is an XDNA file with an optional annotation section

        # Skip the overhangs as I don't know how to represent
        # them in the SeqRecord model...
        _read_overhang(handle)  # right-side overhang
        _read_overhang(handle)  # left-side overhang

        # Read the features
        num_features = unpack('>B', _read(handle, 1))[0]
        while num_features > 0:
            _read_feature(handle, record)
            num_features -= 1

    yield record



class XdnaWriter(SequenceWriter):

    def write_file(self, records):

        if len(records) != 1:
            raise ValueError("A DNA Strider file can only contain one record")

        record = records[0]

        alptype = type(record.seq.alphabet)
        if issubclass(alptype, Alphabet.DNAAlphabet):
            seqtype = 1
        elif issubclass(alptype, Alphabet.RNAAlphabet):
            seqtype = 3
        elif issubclass(alptype, Alphabet.ProteinAlphabet):
            seqtype = 4
        else:
            seqtype = 0

        if record.annotations.get('topology', 'linear') == 'circular':
            topology = 1
        else:
            topology = 0

        # Write header
        self.handle.write(pack('>BBB25xII60xI11xB',
            0, # version
            seqtype, topology, len(record),
            0, # negative length
            len(record.description),
            255 # end of header
            ))

        # Actual sequence and comment
        self.handle.write(str(record.seq))
        self.handle.write(record.description)

        self.handle.write(pack('>B', 0))    # Annotation section marker
        self._write_pstring('0')            # right-side overhang
        self._write_pstring('0')            # left-side overhand

        # Write features
        self.handle.write(pack('>B', len(record.features)))
        for feature in record.features:

            if type(feature.location.start) != ExactPosition or type(feature.location.end) != ExactPosition:
                # Cannot store fuzzy locations, skip feature
                continue

            self._write_pstring(feature.qualifiers.get('label', [''])[0])

            description = ''
            for qname in feature.qualifiers:
                if qname in ('label', 'translation'):
                    continue

                for val in feature.qualifiers[qname]:
                    if len(description) > 0:
                        description = description + '\x0D'
                    description = description + '%s="%s"' % (qname, val)
            self._write_pstring(description)

            self._write_pstring(feature.type)

            start = feature.location.start.position + 1 # 1-based coordinates
            end = feature.location.end.position
            strand = 1
            if feature.location.strand == -1:
                start, end = end, start
                strand = 0
            self._write_pstring(str(start))
            self._write_pstring(str(end))

            self.handle.write(pack('>BBBB', strand, 1, 0, 1))
            self._write_pstring('127,127,127')

        return 1


    def _write_pstring(self, s):
        self.handle.write(pack('>B', len(s)))
        self.handle.write(s)

==> incenp.binseqs-0.2.1.tar.gz: incenp.binseqs-0.2.1/incenp/bio/seqio/XdnaIO.py <==
# -*- coding: utf-8 -*-
# Copyright © 2017,2018 Damien Goutte-Gattat
#
# Redistribution and use of this script, with or without modifications,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of this script must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Xdna format support for Biopython

Provide read/write support for the binary format used by DNA Strider
and SerialCloner.
"""

from re import match
from struct import pack, unpack

from Bio import Alphabet
from Bio.Seq import Seq
from Bio.SeqIO.Interfaces import SequenceWriter
from Bio.SeqFeature import SeqFeature, FeatureLocation, ExactPosition
from Bio.SeqRecord import SeqRecord


_seq_types = {
    0: Alphabet.generic_alphabet,
    1: Alphabet.generic_dna,
    2: Alphabet.generic_dna,
    3: Alphabet.generic_rna,
    4: Alphabet.generic_protein
}

_seq_topologies = {
    0: 'linear',
    1: 'circular'
}


def _read(handle, length):
    data = handle.read(length)
    if len(data) < length:
        raise ValueError("Cannot read %d bytes from handle" % length)
    return data


def _read_pstring(handle):
    length = unpack('>B', _read(handle, 1))[0]
    return unpack('%ds' % length, _read(handle, length))[0].decode('ASCII')


def _read_pstring_as_integer(handle):
    return int(_read_pstring(handle))


def _read_overhang(handle):
    # An overhang is represented in a XDNA file as:
    # - a Pascal string containing the text representation of the
    #   overhang length (3' overhang if negative, 5' overhang if
    #   positive, no overhang if zero);
    # - the actual overhang sequence.
    length = _read_pstring_as_integer(handle)
    if length != 0:
        overhang = _read(handle, abs(length))
        return (length, overhang)
    else:
        return (None, None)


def _parse_feature_description(desc, qualifiers):
    # The 'description' field of a feature sometimes contains
    # several GenBank-like qualifiers, separated by CR characters.
    for part in [x for x in desc.split('\x0D') if len(x) > 0]:
        m = match('^([^=]+)="([^"]+)"?$', part)
        if m:
            qual, value = m.groups()
            qualifiers[qual] = [value]
        elif not '"' in part:   # Reject ill-formed qualifiers
            qualifiers['note'] = [part]


def _read_feature(handle, record):
    name = _read_pstring(handle)
    desc = _read_pstring(handle)
    type = _read_pstring(handle) or 'misc_feature'
    start = _read_pstring_as_integer(handle)
    end = _read_pstring_as_integer(handle)

    # Feature flags (4 bytes):
    # byte 1 is the strand (0: reverse strand, 1: forward strand);
    # byte 2 tells whether to display the feature;
    # byte 4 tells whether to draw an arrow when displaying the feature;
    # meaning of byte 3 is unknown.
    (forward, display, arrow) = unpack('>BBxB', _read(handle, 4))
    if forward:
        strand = 1
    else:
        strand = -1
        start, end = end, start

    # The last field is a Pascal string usually containing a
    # comma-separated triplet of numbers ranging from 0 to 255.
    # I suspect this represents the RGB color to use when displaying
    # the feature. Skip it as we have no need for it.
    _read_pstring(handle)

    # Assemble the feature
    # Shift start by -1 as XDNA feature coordinates are 1-based
    # while Biopython uses 0-based couting.
    location = FeatureLocation(start - 1, end, strand=strand)
    qualifiers = {}
    if name:
        qualifiers['label'] = [name]
    _parse_feature_description(desc, qualifiers)
    feature = SeqFeature(location, type=type, qualifiers=qualifiers)
    record.features.append(feature)


def XdnaIterator(handle):

    # Parse fixed-size header and do some rudimentary checks
    header = _read(handle, 112)
    (version, type, topology, length, neg_length, com_length) = \
            unpack('>BBB25xII60xI12x', header)
    if version != 0:
        raise ValueError("Unsupported XDNA version")
    if not type in _seq_types:
        raise ValueError("Unknown sequence type")

    # Read actual sequence and comment found in all XDNA files
    sequence = _read(handle, length).decode('ASCII')
    comment = _read(handle, com_length).decode('ASCII')

    # Try to derive a name from the first "word" of the comment.
    name = comment.split(' ')[0]

    # Create record object
    record = SeqRecord(Seq(sequence, _seq_types[type]),
            description=comment, name=name, id=name)
    if topology in _seq_topologies:
        record.annotations['topology'] = _seq_topologies[topology]

    if len(handle.read(1)) == 1:
        # This is an XDNA file with an optional annotation section

        # Skip the overhangs as I don't know how to represent
        # them in the SeqRecord model...
        _read_overhang(handle)  # right-side overhang
        _read_overhang(handle)  # left-side overhang

        # Read the features
        num_features = unpack('>B', _read(handle, 1))[0]
        while num_features > 0:
            _read_feature(handle, record)
            num_features -= 1

    yield record



class XdnaWriter(SequenceWriter):

    def write_file(self, records):

        if len(records) != 1:
            raise ValueError("A DNA Strider file can only contain one record")

        record = records[0]

        alptype = type(record.seq.alphabet)
        if issubclass(alptype, Alphabet.DNAAlphabet):
            seqtype = 1
        elif issubclass(alptype, Alphabet.RNAAlphabet):
            seqtype = 3
        elif issubclass(alptype, Alphabet.ProteinAlphabet):
            seqtype = 4
        else:
            seqtype = 0

        if record.annotations.get('topology', 'linear') == 'circular':
            topology = 1
        else:
            topology = 0

        # We store the record's id and description in the comment field.
        # Make sure to avoid duplicating the id if it is already
        # contained in the description.
        if record.description.startswith(record.id):
            comment = record.description
        else:
            comment = '{} {}'.format(record.id, record.description)

        # Write header
        self.handle.write(pack('>BBB25xII60xI11xB',
            0, # version
            seqtype, topology, len(record),
            0, # negative length
            len(comment),
            255 # end of header
            ))

        # Actual sequence and comment
        self.handle.write(str(record.seq).encode('ASCII'))
        self.handle.write(comment.encode('ASCII'))

        self.handle.write(pack('>B', 0))    # Annotation section marker
        self._write_pstring('0')            # right-side overhang
        self._write_pstring('0')            # left-side overhand

        # Write features
        # We must skip features with fuzzy locations as they cannot be
        # represented in the Xdna format
        features = [f for f in record.features if type(f.location.start) == ExactPosition and type(f.location.end) == ExactPosition]
        # We also cannot store more than 255 features as the number of
        # features is stored on a single byte...
        if len(features) > 255:
            features = features[:255]
        self.handle.write(pack('>B', len(features)))
        for feature in features:
            self._write_pstring(feature.qualifiers.get('label', [''])[0])

            description = ''
            for qname in feature.qualifiers:
                if qname in ('label', 'translation'):
                    continue

                for val in feature.qualifiers[qname]:
                    if len(description) > 0:
                        description = description + '\x0D'
                    description = description + '%s="%s"' % (qname, val)
            self._write_pstring(description)

            self._write_pstring(feature.type)

            start = feature.location.start.position + 1 # 1-based coordinates
            end = feature.location.end.position
            strand = 1
            if feature.location.strand == -1:
                start, end = end, start
                strand = 0
            self._write_pstring(str(start))
            self._write_pstring(str(end))

            self.handle.write(pack('>BBBB', strand, 1, 0, 1))
            self._write_pstring('127,127,127')

        return 1


    def _write_pstring(self, s):
        if len(s) > 255:
            s = s[:255]
        self.handle.write(pack('>B', len(s)))
        self.handle.write(s.encode('ASCII'))
